// Title:	BiDiPro DLL module
// Desc:	SPI DataFlash
// Ver:		0.0.2
// Found:	2011-05-25
// Last:	2012-08-28
// CPU:		386+
// Model:	Tiny
// Compil:	Borland C++ 3.1 (DOS)
// Author:	Oleg O. Chukaev aka 0leg


/****
AT45 BiDiPro module
Copyright (c) 2011-2012, 㪠   <oleg.chukaev@mail.ru>

蠥 ୮ ࠭  ᯮ짮    室
, ⠪   筮 ଥ,  ﬨ  ,  ᮡ ᫥
᫮:

1.  ୮ ࠭ 室   ⠢ 㪠
    㢥  ᪮ ࠢ,  ᯨ᮪ ᫮  ᫥騩
   ⪠  ࠭⨩.
2.  ୮ ࠭ 筮   ࠭ 㪠
    ଠ  ᪮ ࠢ,  ᯨ᮪ ᫮  ᫥騩
   ⪠  ࠭⨩  㬥樨 /  㣨 ਠ, ⠢塞
    ࠭.

      /  
"  "  -  ,   
, ,    ,  
      .    ,
    ,      ,
         ,   
/   ,    ,  
,   , ,   
,      
(,     ,  , 
,    -    , 
      ),   
         .
****/


#include <stdlib.h>
#include "../dlldata.h"
#include "../errors.h"
#include "../bidi.h"


#ifdef LINUXXX
 extern DLLDATA *dlp;
#else
# define dlp ((DLLDATA *) NULL)		/* ⥫    */
#endif

#define bdport dlp->bdprt		/* ⥫  hardware ࠩ */

typedef struct CHDEFS {
	dword size;	// 𬪮
	word page;	//  ࠭
	word s1;	// 饭  ࠧ 
	word s2;	//  ࠭栬 ⮢
} CHDEFS;

CHDEFS chd[]={
	{0x021000L, 0x108, 1, 7},		//011
	{0x042000L, 0x108, 1, 7},		//021
	{0x084000L, 0x108, 1, 7},		//041
	{0x210000L, 0x108, 1, 7},		//081
	{0x420000L, 0x210, 2, 6},		//161
	{0x840000L, 0x210, 2, 6},		//321
};

// ७   CFG:
static struct CFGKEY cfgkeys[] = {
	{'D', "wtime", 20000, 20000, 1000000L},   //Write time (us)
	{'D', "betime", 10000, 10000, 40000000L}, //Block erase time (us)
	{'D', "rb_pin", 0, 0, 1},                 //Use RDY/-BY pin? (1 = yes)
	{0, {0}, 0, 0},                           //End of table
};

#define cfg_wtime	cfgkeys[0].data
#define cfg_betime	cfgkeys[1].data
#define cfg_rb_pin	cfgkeys[2].data

/* Output, 1B */
#define PIN_CS		0x01
#define PIN_WP		0x04
#define PIN_SI		0x08
#define PIN_SCK		0x10
#define PIN_RESET	0x20

/* Input, 1A */
#define PIN_SO		0x01
#define PIN_RB		0x10

/* DataFlash commands */
#define DF_READ		0xe8
#define DF_PGERASE	0x81
#define DF_BLKERASE	0x50
#define DF_RDSR		0xd7
#define DF_PRGTHRBUF	0x82
#define DF_BUFWRITE	0x84
#define DF_PRGWOERASE	0x88

#define	BUSY_TIMEOUT	500000L


static char *copyright = "AT45 0.0.2  SPI DataFlash module  "
	"(c) Oleg O. Chukaev  ["__DATE__"]";
static word chtype = 0xffff;
static CHDEFS *pchd;
static byte pin_wp = 0;


/*
 | 몫祭 ⠭
 */
static void power_off (void) {
	bdport (PRST);
}


/*
 | 祭 ⠭
 */
static int power_on (void) {

	if (chtype == 0xffff)
		return (ERC_CHIPUNSP);

	bdport (PFORCE | PWR1 | PAIN);		/* DD1:IOO */
	bdport (PFORCE | PWR2);			/* DD4:OOO */
	dlp->voltage (dlp->Vcc, 0, 0);
	bdport (PWB2 | 0x80);			/* Vcc pin 36 */
	bdport (PWA2 | BITTER);
	bdport (PWB1 | PIN_RESET | pin_wp | PIN_CS);
	dlp->mcsdelay (10000);
	return (0);
}


/*
 | Get value of one pin
 |
 | Ret: 0 / 1
 */
static byte get_pin (byte pin) {
	return (bdport (PRA1) & pin) ? 1 : 0;
}


/*
 | Set all pins
 */
static void set_pins (byte pins) {
	bdport (PWB1 | pins);
}


static void pin_ce_1 (void) {
  	set_pins (PIN_RESET | pin_wp | PIN_CS);
}


static void pin_ce_0 (void) {
  	set_pins (PIN_RESET | pin_wp);
}


/*
 |  
 */
static byte spi_read_byte (void) {

	word i;
	byte d = 0;

	for (i = 0; i < 8; i++) {
		d <<= 1;
		d |= get_pin (PIN_SO);
		set_pins (pin_wp | PIN_RESET | PIN_SCK);/* SCK=1 */
		set_pins (pin_wp | PIN_RESET);		/* SCK=0 */
	}
	return d;
}


/*
 | । 
 */
static int spi_write_byte (byte d) {

	word i;
	byte z;

	for (i = 0; i < 8; i++) {
		z = (d & 0x80) ? PIN_SI : 0;
		set_pins (pin_wp | PIN_RESET | z);		/* data */
		set_pins (pin_wp | PIN_RESET | PIN_SCK | z);	/* SCK=1 */
		set_pins (pin_wp | PIN_RESET);			/* SCK=0 */
		d <<= 1;
	}
	return 0;
}


/*
 | Poll RDY/-BUSY bit or pin
 */
static int poll_busy_bit (void) {

	dword i;

	for (i = 0; i < BUSY_TIMEOUT / 20; i++) {
		dlp->mcsdelay (20);
		if (cfg_rb_pin) {
			if (get_pin (PIN_RB)) {
				pin_ce_1 ();
				return 0;
			}
		} else {
			pin_ce_1 ();
			spi_write_byte (DF_RDSR);
			if ((spi_read_byte () & 0x80) == 0x80) {
				pin_ce_1 ();
				return 0;
			}
		}
	}
	pin_ce_1 ();
	return 1;
}


/*
 | Read, compare, blank check
 */
static int read_all (void) {

	dword adr;
	dword istep = pchd->size / 64L;
	dword errors = 0;
	byte dat;

	power_on ();
	pin_ce_0 ();
	set_pins (PIN_RESET);
	spi_write_byte (DF_READ);
	spi_write_byte (0x00);		/* A23..A16 */
	spi_write_byte (0x00);		/* A15..A8 */
	spi_write_byte (0x00);		/* A7..A0 */
	spi_write_byte (0x00);		/* Don't care */
	spi_write_byte (0x00);		/* --//-- */
	spi_write_byte (0x00);		/* --//-- */
	spi_write_byte (0x00);		/* --//-- */
	for (adr = 0; adr < pchd->size; adr++) {
		dat = spi_read_byte ();
		switch (dlp->funcode) {
		case DFN_FREAD:
			dlp->setbytebuf (adr, dat);
			break;
		case DFN_FCOMPARE:
			if (dlp->getbytebuf (adr) != dat) {
				/* ⬥ ᮢ */
				dlp->setbmark (adr, 1);
				errors++;
			}
			break;
		case DFN_FBLANK:
			if (dat != 0xff) {
				dlp->adr = adr;
				power_off ();
				return ERC_BLANK;
			}
		}
		if (adr % istep == 0 && dlp->iprogress (adr, 2))
			break;
	}
	power_off ();
	dlp->edata = errors;
	return 0;
}


/*
 | Read Status Register
 */
static int read_id (void) {

	dword id1 = 0, id2 = 0;

	power_on ();
	pin_ce_0 ();
	spi_write_byte (DF_RDSR);
	id2 = spi_read_byte ();
	pin_ce_1 ();
	power_off ();
	dlp->edata = id1 | (id2 << 16L);
	return 0;
}


/*
 | Erase
 */
static int erase (void) {

	dword adr;
	dword istep = pchd->size / 64L;

	dlp->iprogress (pchd->size, 0);
	pin_wp = PIN_WP;
	power_on ();
	pin_ce_0 ();
	for (adr = 0; adr < pchd->size / (pchd->page * 8); adr++) {
		spi_write_byte (DF_BLKERASE);
		spi_write_byte ((adr >> pchd->s2-3) & 0xff); /* PAx:PA7 */
		spi_write_byte ((adr << pchd->s1+3) & 0xff); /* PA6:PA3,X,X,X,X */
		spi_write_byte (0);			     /* 8 * X */
		pin_ce_1 ();
		if (cfg_betime != 0) {
			/* Fixed delay */
			dlp->mcsdelay (cfg_betime);
		} else {
			/* Poll RDY/-BUSY bit */
			if (poll_busy_bit ()) {
				power_off ();
				pin_wp = 0;
				dlp->adr = adr * pchd->page;
				return ERC_WRITE;
			}
		}
		pin_ce_0 ();
		if ((adr * pchd->page * 8) % istep == 0
				&& dlp->iprogress (adr * pchd->page * 8, 2)) {
			break;
		}
	}
	pin_ce_1 ();
	power_off ();
	dlp->iprogress (0, 1);				/* close */
	pin_wp = 0;
	return 0;
}


/*
 | Write
 */
static int write_all (void) {

	dword i;
	dword adr = 0;
	dword istep = pchd->size / 64L;

	pin_wp = PIN_WP;
	power_on ();
	pin_ce_0 ();
	for (adr = 0; adr < pchd->size / pchd->page; adr++) {
		spi_write_byte (DF_PRGTHRBUF);
		spi_write_byte ((adr >> pchd->s2) & 0xff); /* PAx:PA7 */
		spi_write_byte ((adr << pchd->s1) & 0xff); /* PA6:PA0,BFA8 */
		spi_write_byte (0);			   /* BFA7:BFA0 */
		for (i = 0; i < pchd->page; i++) {
			spi_write_byte (dlp->getbytebuf (
				adr * pchd->page + i));
		}
		pin_ce_1 ();
		if (cfg_wtime != 0) {
			/* Fixed delay */
			dlp->mcsdelay (cfg_wtime);
		} else {
			/* Poll RDY/-BUSY bit */
			if (poll_busy_bit ()) {
				power_off ();
				pin_wp = 0;
				dlp->adr = adr * pchd->page;
				return ERC_WRITE;
			}
		}
		pin_ce_0 ();
		if ((adr * pchd->page) % istep == 0
				&& dlp->iprogress (adr * pchd->page, 2)) {
			break;
		}
	}
	pin_ce_1 ();
	power_off ();
	pin_wp = 0;
	return 0;
}


/*
 | Write w/o erasing
 */
static int test (void) {

	dword i;
	dword adr = 0;
	dword istep = pchd->size / 64L;

	dlp->iprogress (pchd->size, 0);
	pin_wp = PIN_WP;
	power_on ();
	pin_ce_0 ();
	for (adr = 0; adr < pchd->size / pchd->page; adr++) {
		spi_write_byte (DF_BUFWRITE);
		spi_write_byte (0);			/* X */
		spi_write_byte (0);			/* X, (BFA9), BFA8 */
		spi_write_byte (0);			/* BFA7:BFA0 */
		for (i = 0; i < pchd->page; i++) {
			spi_write_byte (dlp->getbytebuf (
				adr * pchd->page + i));
		}
		pin_ce_1 ();
		pin_ce_0 ();
		spi_write_byte (DF_PRGWOERASE);
		spi_write_byte ((adr >> pchd->s2) & 0xff); /* PAx:PA7 */
		spi_write_byte ((adr << pchd->s1) & 0xff); /* PA6:PA0,BFA8 */
		spi_write_byte (0);			   /* BFA7:BFA0 */
		pin_ce_1 ();
		if (cfg_wtime != 0) {
			/* Fixed delay */
			dlp->mcsdelay (cfg_wtime);
		} else {
			/* Poll RDY/-BUSY bit */
			if (poll_busy_bit ()) {
				power_off ();
				pin_wp = 0;
				dlp->adr = adr * pchd->page;
				return ERC_WRITE;
			}
		}
		pin_ce_0 ();
		if ((adr * pchd->page) % istep == 0
				&& dlp->iprogress (adr * pchd->page, 2)) {
			break;
		}
	}
	pin_ce_1 ();
	power_off ();
	dlp->iprogress (0, 1);				/* close */
	pin_wp = 0;
	return 0;
}


/*
 | ⠭ ⨯ ROM
 */
static word settype (void) {

	word i;

	for (i = 0; i < sizeof(chd) / sizeof(CHDEFS); i++) {
		if (chd[i].size == dlp->esize) {
			return (chtype = i);
		}
	}
	return (chtype = 0xffff);
}


/*
 | Main code
 */
int dmain (void) {

	switch (dlp->funcode) {
	case DFN_LOAD:
		/* ⨢஢ 㯭 㭪樨: */
		dlp->funmask[DFN_INIT] = 1;
		dlp->funmask[DFN_FREAD] = 1;
		dlp->funmask[DFN_FWRITE] = 1;
		dlp->funmask[DFN_FCOMPARE] = 1;
		dlp->funmask[DFN_READID] = 1;
		dlp->funmask[DFN_ERASE] = 1;
		dlp->funmask[DFN_FBLANK] = 1;
		dlp->funmask[DFN_TEST] = 1;
		dlp->funmask[DFN_LOCK] = 0;
		dlp->defstate = 0xff;
		dlp->copyright = (byte *)copyright;
		dlp->version = DLLVERSION;
		dlp->cfgkeys = cfgkeys;
		return 0;

	case DFN_INIT:
		if (settype () == 0xffff)
			return ERC_CHIPUNSP;
		pchd = &chd[chtype];
		dlp->dippins = 8;
		return 0;

	case DFN_FREAD:
		return read_all ();

	case DFN_FWRITE:
		return write_all ();

	case DFN_FCOMPARE:
		return read_all ();

	case DFN_FBLANK:
		return read_all ();

	case DFN_READID:
		return read_id ();

	case DFN_ERASE:
		return erase ();

	case DFN_TEST:
		return test ();

	default:
		return ERC_FUNUNSP;
	}
}

/* E0F */
